#define _GNU_SOURCE
#include <math.h>
#include <err.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <sys/time.h>
#include <time.h>
#include <limits.h>
#include "stdrusty.h"
#include "list/list.h"
#include <SDL/SDL.h>
#include <cwiid.h>
#include <assert.h>

#define MAX_X CWIID_IR_X_MAX
#define MAX_Y CWIID_IR_Y_MAX

struct coord
{
	double x;
	double y;
};

/* Obtained from http://tog.acm.org/GraphicsGems/gemsiii/insectc.c */
/* Faster Line Segment Intersection   */
/* Franklin Antonio                   */
static bool lines_intersect(const struct coord *start1,
			    const struct coord *start2,
			    const struct coord *start3,
			    const struct coord *start4,
			    struct coord *intersect)
{
	double Ax,Bx,Cx,Ay,By,Cy,d,e,f,num;
	struct coord lo, hi;

	Ax = start2->x - start1->x;
	Bx = start3->x - start4->x;

	if(Ax<0) {					/* X bound box test*/
		lo.x = start2->x; hi.x = start1->x;
	} else {
		hi.x = start2->x; lo.x = start1->x;
	}
	if(Bx>0) {
		if(hi.x < start4->x || start3->x < lo.x) return false;
	} else {
		if(hi.x < start3->x || start4->x < lo.x) return false;
	}

	Ay = start2->y - start1->y;
	By = start3->y - start4->y;

	if(Ay<0) {					/* Y bound box test*/
		lo.y = start2->y; hi.y = start1->y;
	} else {
		hi.y = start2->y; lo.y = start1->y;
	}
	if(By>0) {
		if(hi.y < start4->y || start3->y < lo.y) return false;
	} else {
		if(hi.y < start3->y || start4->y < lo.y) return false;
	}


	Cx = start1->x - start3->x;
	Cy = start1->y - start3->y;
	d = By*Cx - Bx*Cy;				/* alpha numerator*/
	f = Ay*Bx - Ax*By;				/* both denominator*/
	if (f>0) {					/* alpha tests*/
		if (d < 0 || d > f) return false;
	} else {
		if (d > 0 || d < f) return false;
	}

	e = Ax*Cy - Ay*Cx;				/* beta numerator*/
	if (f > 0) {					/* beta tests*/
		if (e < 0 || e > f) return false;
	} else {
		if (e > 0 || e < f) return false;
	}

/*compute intersection coordinates*/
	/* Don't worry about tiny values of f (< 1/2 pixel across screen) */
	if (fabs(f) < 1.0 / (2.0 * MAX_X * MAX_Y))
		return false;

	num = d*Ax;					/* numerator */
	intersect->x = start1->x + num / f;		/* intersection x */

	num = d*Ay;
	intersect->y = start1->y + num / f;		/* intersection y */

	return true;
}

// A set of very useful macros that you will find in most
// code that I write whether I use them in a program or
// not.

#define abs(a) (((a)<0) ? -(a) : (a))
#define sign(a) (((a)<0) ? -1 : (a)>0 ? 1 : 0)

// The following code implements a Bresenham line drawing
// algorithm. There are 4 separate routines each optimized
// for one of the four pixel depths supported by SDL. SDL
// support many pixel formats, but it only support 8, 16,
// 24, and 32 bit pixels.

//----------------------------------------------------------

// Draw lines in 8 bit surfaces.

static void line8(SDL_Surface *s, 
                  int x1, int y1, 
                  int x2, int y2, 
                  Uint32 color)
{
  int d;
  int x;
  int y;
  int ax;
  int ay;
  int sx;
  int sy;
  int dx;
  int dy;

  Uint8 *lineAddr;
  Sint32 yOffset;

  dx = x2 - x1;  
  ax = abs(dx) << 1;  
  sx = sign(dx);

  dy = y2 - y1;  
  ay = abs(dy) << 1;  
  sy = sign(dy);
  yOffset = sy * s->pitch;

  x = x1;
  y = y1;

  lineAddr = ((Uint8 *)(s->pixels)) + (y * s->pitch);
  if (ax>ay)
  {                      /* x dominant */
    d = ay - (ax >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *(lineAddr + x) = (Uint8)color;

      if (x == x2)
      {
        return;
      }
      if (d>=0)
      {
        y += sy;
        lineAddr += yOffset;
        d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else
  {                      /* y dominant */
    d = ax - (ay >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *(lineAddr + x) = (Uint8)color;

      if (y == y2)
      {
        return;
      }
      if (d>=0) 
      {
        x += sx;
        d -= ay;
      }
      y += sy;
      lineAddr += yOffset;
      d += ax;
    }
  }
}

//----------------------------------------------------------

// Draw lines in 16 bit surfaces. Note that this code will
// also work on 15 bit surfaces.

static void line16(SDL_Surface *s, 
                   int x1, int y1, 
                   int x2, int y2, 
                   Uint32 color)
{
  int d;
  int x;
  int y;
  int ax;
  int ay;
  int sx;
  int sy;
  int dx;
  int dy;

  Uint8 *lineAddr;
  Sint32 yOffset;

  dx = x2 - x1;  
  ax = abs(dx) << 1;  
  sx = sign(dx);

  dy = y2 - y1;  
  ay = abs(dy) << 1;  
  sy = sign(dy);
  yOffset = sy * s->pitch;

  x = x1;
  y = y1;

  lineAddr = ((Uint8 *)s->pixels) + (y * s->pitch);
  if (ax>ay)
  {                      /* x dominant */
    d = ay - (ax >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *((Uint16 *)(lineAddr + (x << 1))) = (Uint16)color;

      if (x == x2)
      {
        return;
      }
      if (d>=0)
      {
        y += sy;
        lineAddr += yOffset;
        d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else
  {                      /* y dominant */
    d = ax - (ay >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *((Uint16 *)(lineAddr + (x << 1))) = (Uint16)color;

      if (y == y2)
      {
        return;
      }
      if (d>=0) 
      {
        x += sx;
        d -= ay;
      }
      y += sy;
      lineAddr += yOffset;
      d += ax;
    }
  }
}

//----------------------------------------------------------

// Draw lines in 24 bit surfaces. 24 bit surfaces require
// special handling because the pixels don't fall on even
// address boundaries. Instead of being able to store a
// single byte, word, or long you have to store 3
// individual bytes. As a result 24 bit graphics is slower
// than the other pixel sizes.

static void line24(SDL_Surface *s, 
                   int x1, int y1, 
                   int x2, int y2, 
                   Uint32 color)
{
  int d;
  int x;
  int y;
  int ax;
  int ay;
  int sx;
  int sy;
  int dx;
  int dy;

  Uint8 *lineAddr;
  Sint32 yOffset;

#if (SDL_BYTEORDER == SDL_BIG_ENDIAN)
  color <<= 8;
#endif

  dx = x2 - x1;  
  ax = abs(dx) << 1;  
  sx = sign(dx);

  dy = y2 - y1;  
  ay = abs(dy) << 1;  
  sy = sign(dy);
  yOffset = sy * s->pitch;

  x = x1;
  y = y1;

  lineAddr = ((Uint8 *)(s->pixels)) + (y * s->pitch);
  if (ax>ay)
  {                      /* x dominant */
    d = ay - (ax >> 1);
    for (;;)
    {
      Uint8 *p = (lineAddr + (x * 3));
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        memcpy(p, &color, 3);

      if (x == x2)
      {
        return;
      }
      if (d>=0)
      {
        y += sy;
        lineAddr += yOffset;
        d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else
  {                      /* y dominant */
    d = ax - (ay >> 1);
    for (;;)
    {
      Uint8 *p = (lineAddr + (x * 3));
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        memcpy(p, &color, 3);

      if (y == y2)
      {
        return;
      }
      if (d>=0) 
      {
        x += sx;
        d -= ay;
      }
      y += sy;
      lineAddr += yOffset;
      d += ax;
    }
  }
}

//----------------------------------------------------------

// Draw lines in 32 bit surfaces. Note that this routine
// ignores alpha values. It writes them into the surface
// if they are included in the pixel, but does nothing
// else with them.

static void line32(SDL_Surface *s, 
                   int x1, int y1, 
                   int x2, int y2, 
                   Uint32 color)
{
  int d;
  int x;
  int y;
  int ax;
  int ay;
  int sx;
  int sy;
  int dx;
  int dy;

  Uint8 *lineAddr;
  Sint32 yOffset;

  dx = x2 - x1;  
  ax = abs(dx) << 1;  
  sx = sign(dx);

  dy = y2 - y1;  
  ay = abs(dy) << 1;  
  sy = sign(dy);
  yOffset = sy * s->pitch;

  x = x1;
  y = y1;

  lineAddr = ((Uint8 *)(s->pixels)) + (y * s->pitch);
  if (ax>ay)
  {                      /* x dominant */
    d = ay - (ax >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *((Uint32 *)(lineAddr + (x << 2))) = (Uint32)color;

      if (x == x2)
      {
        return;
      }
      if (d>=0)
      {
        y += sy;
        lineAddr += yOffset;
        d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else
  {                      /* y dominant */
    d = ax - (ay >> 1);
    for (;;)
    {
      if (y >= 0 && y < MAX_Y && x >= 0 && x < MAX_X)
        *((Uint32 *)(lineAddr + (x << 2))) = (Uint32)color;

      if (y == y2)
      {
        return;
      }
      if (d>=0) 
      {
        x += sx;
        d -= ay;
      }
      y += sy;
      lineAddr += yOffset;
      d += ax;
    }
  }
}

//----------------------------------------------------------

// Examine the depth of a surface and select a line
// drawing routine optimized for the bytes/pixel of the
// surface.

static void line(SDL_Surface *s, 
                 int x1, int y1, 
                 int x2, int y2, 
                 Uint32 color)
{
  switch (s->format->BytesPerPixel)
  {
  case 1:
    line8(s, x1, y1, x2, y2, color);
    break;
  case 2:
    line16(s, x1, y1, x2, y2, color);
    break;
  case 3:
    line24(s, x1, y1, x2, y2, color);
    break;
  case 4:
    line32(s, x1, y1, x2, y2, color);
    break;
  }
}

struct ball
{
	struct coord pos;
	struct coord move;
};

struct line_segment
{
	struct coord start, end;
	struct list_node list;
	bool ignore;
	struct timeval expires;
};

static void thick_line(SDL_Surface *s, const struct line_segment *l,
		       unsigned color)
{
	struct line_segment adj;
	int x, y, minx = MAX_X, miny = MAX_Y, maxx = 0, maxy = 0;

	SDL_LockSurface(s);

	/* Cheap hack */
	for (x = -1; x < 2; x++) {
		for (y = -1; y < 2; y++) {
			adj.start.x = l->start.x + x;
			adj.end.x = l->end.x + x;
			adj.start.y = l->start.y + y;
			adj.end.y = l->end.y + y;

			if (adj.start.x < minx)
				minx = adj.start.x;
			if (adj.start.x > maxx)
				maxx = adj.start.x;
			if (adj.start.y < miny)
				miny = adj.start.y;
			if (adj.start.y > maxy)
				maxy = adj.start.y;
			if (adj.end.x < minx)
				minx = adj.end.x;
			if (adj.end.x > maxx)
				maxx = adj.end.x;
			if (adj.end.y < miny)
				miny = adj.end.y;
			if (adj.end.y > maxy)
				maxy = adj.end.y;

			line(s, adj.start.x, adj.start.y, adj.end.x, adj.end.y,
			     color);
		}
	}

	if (minx < 0)
		minx = 0;
	if (miny < 0)
		miny = 0;
	if (maxx >= MAX_X)
		maxx = MAX_X-1;
	if (maxy >= MAX_Y)
		maxy = MAX_Y-1;
	SDL_UpdateRect(s, minx, miny, maxx-minx, maxy-miny);
	SDL_UnlockSurface(s);
}

static bool intersect(const struct coord *c1, const struct coord *c2,
		      const struct line_segment *seg, struct coord *intersect)
{
	return lines_intersect(c1, c2, &seg->start, &seg->end, intersect);
}

static double dist(const struct coord *c1, const struct coord *c2)
{
	double x = (c1->x - c2->x), y = (c1->y - c2->y);
	return sqrt(x * x + y * y);
}

static unsigned rad_to_deg(double rad)
{
	return ((unsigned)(rad / M_PI * 180) + 360) % 360;
}

static void bounce(struct ball *ball, const struct line_segment *line,
		   struct coord *move, double increase_speed)
{
	double len, speed, lang, iang, oang;
	struct coord isect, new;

	new.x = ball->pos.x + move->x;
	new.y = ball->pos.y + move->y;

	/* How far were we supposed to move ball? */
	len = sqrt(move->x * move->x + move->y * move->y);

	/* How far is it to intersection? */
	if (!intersect(&ball->pos, &new, line, &isect))
		errx(1, "No intersection any more?\n");;
	len -= dist(&ball->pos, &isect);

	/* Move ball to intersection. */
	ball->pos = isect;
	printf("ball is now at %f,%f\n", ball->pos.x, ball->pos.y);

	/* Outgoing angle = 2 * line angle - incident angle. */
	lang = atan2(line->end.x - line->start.x, line->end.y - line->start.y);
	iang = atan2(ball->move.x, ball->move.y);
	oang = 2 * lang - iang, 2*M_PI;
	printf("lang = %u, iang = %u, oang=%u\n", 
	       rad_to_deg(lang),
	       rad_to_deg(iang),
	       rad_to_deg(oang));

	/* Set new direction for ball, at slightly increased speed */
	speed = sqrt(ball->move.x * ball->move.x + ball->move.y * ball->move.y);
	speed += increase_speed;
	ball->move.x = sin(oang) * speed;
	ball->move.y = cos(oang) * speed;

	/* Set move. */
	move->x = sin(oang) * len;
	move->y = cos(oang) * len;
	printf("len = %f, move = %f,%f\n", len, move->x, move->y);
}

static struct line_segment border[] = {
	{ { 0, 0, }, { MAX_X-1, 0 } },
	{ { MAX_X-1, 0, }, { MAX_X-1, MAX_Y-1 } },
	{ { MAX_X-1, MAX_Y-1, }, { 0, MAX_Y-1 } },
	{ { 0, MAX_Y-1, }, { 0, 0 } },
};

static inline float deg_to_rad(float degrees)
{
	return degrees * M_PI / 180;
}

static SDL_Surface *sub_surface(SDL_Surface *screen, int w, int h)
{
	return SDL_CreateRGBSurface(SDL_SWSURFACE, w, h,
				    screen->format->BitsPerPixel,
				    screen->format->Rmask,
				    screen->format->Gmask,
				    screen->format->Bmask,
				    screen->format->Amask);
}

static SDL_Surface *ball_surface(SDL_Surface *screen)
{
	SDL_Surface *ball;

	/* Just like the screen surface. */
	ball = sub_surface(screen, 5, 5);

	/* Lock the surface */
	SDL_LockSurface(ball);

	/* Black, white corners. */
	memset(ball->pixels, 0x00, ball->pitch * 5);
	line(ball, 0, 0, 0, 0, 0xFFFFFFFF);
	line(ball, 4, 0, 4, 0, 0xFFFFFFFF);
	line(ball, 0, 4, 0, 4, 0xFFFFFFFF);
	line(ball, 4, 4, 4, 4, 0xFFFFFFFF);
	SDL_UnlockSurface(ball);

	return ball;
}

static void clear_ignore(struct list_head *lines)
{
	struct line_segment *i;
	printf("Unignoring...\n");
	list_for_each(lines, i, list)
		i->ignore = false;
}

int main(int argc, char *argv[])
{
	struct ball ball;
	SDL_Surface *screen, *ballsur, *savesur = NULL;
	Uint8  video_bpp;
	Uint32 videoflags;
	unsigned int i, time_since_last_ir = INT_MAX;
	SDL_Rect rect;
	struct timeval line_life = { 3, 0 };
	cwiid_wiimote_t *wiimote;
	struct cwiid_state state;
	bdaddr_t addr = *BDADDR_ANY;
	LIST_HEAD(lines);
	struct cwiid_ir_src *ir, last_ir = { .valid = 0 };

	videoflags = SDL_SWSURFACE;

	if (argv[1] && streq(argv[1], "--fullscreen")) {
		videoflags = SDL_SWSURFACE;
		argc--;
		argv++;
	}

	if (argc != 5)
		errx(1, "Usage: %s [--fullscreen] ballx bally ballangle ballspeed\n"
		     "   Where ballangle is 0 for north, 90 for east, etc\n",
			argv[0]);

	ball.pos.x = atol(argv[1]);
	ball.pos.y = atol(argv[2]);
	ball.move.x = sin(deg_to_rad(atof(argv[3]))) * atol(argv[4]);
	ball.move.y = cos(deg_to_rad(atof(argv[3]))) * atol(argv[4]);
	printf("ball move = %f,%f\n", ball.move.x, ball.move.y);

	/* Initialize SDL */
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
		return(1);
	}

	video_bpp = 0;

	printf("Put Wiimote in discoverable mode (press 1+2)\n");
	wiimote = cwiid_open(&addr, 0);
	if (!wiimote)
		errx(1, "Can't find the Wiimote");

	if (cwiid_set_rpt_mode(wiimote, CWIID_RPT_IR))
		errx(1, "Can't set IR repeat mode");

	if ( (screen=SDL_SetVideoMode(MAX_X,MAX_Y,video_bpp,videoflags)) == NULL ) {
		errx(1, "Couldn't set %dx%dx%d video mode: %s",
		     MAX_X, MAX_Y,video_bpp, SDL_GetError());
	}

	/* Set the surface pixels and refresh! */
	if ( SDL_LockSurface(screen) < 0 ) {
		errx(1, "Couldn't lock the display surface: %s",
		     SDL_GetError());
	}
	/* White. */
	memset(screen->pixels, 0xFF, screen->pitch * screen->h);

	SDL_UnlockSurface(screen);
	SDL_UpdateRect(screen, 0, 0, 0, 0);

	/* Draw borders, put them in list. */
	for (i = 0; i < ARRAY_SIZE(border); i++) {
		border[i].ignore = false;
		border[i].expires.tv_sec = LONG_MAX;
		border[i].expires.tv_usec = 0;
		list_add(&lines, &border[i].list);
		thick_line(screen, &border[i], 0);
	}

	ballsur = ball_surface(screen);

	for (;;) {
		struct coord new, isect, move = ball.move;
		struct line_segment *best_l, *l, *next;
		double best_d;
		SDL_Event event;
		struct timeval now;
		bool ignored = false;

		gettimeofday(&now, NULL);

		/* Expire lines */
		list_for_each_safe(&lines, l, next, list) {
			if (timercmp(&l->expires, &now, <)) {
				printf("Deleting line %p\n", l);
				list_del(&l->list);
				/* FIXME: Undraw properly. */
				thick_line(screen, l, 0xFFFFFFFF);
				free(l);
			}
		}

	again:
		best_d = MAX_X + MAX_Y;
		best_l = NULL;
		new.x = ball.pos.x + move.x;
		new.y = ball.pos.y + move.y;

		list_for_each(&lines, l, list) {
			if (!l->ignore && intersect(&ball.pos, &new, l, &isect)) {
				double d;
				d = dist(&ball.pos, &isect);
				if (d < best_d) {
					best_l = l;
					best_d = d;
				}
			}
		}

		if (best_l) {
			printf("Ball bouncing off (%f,%f %f,%f)!\n",
			       best_l->start.x, best_l->start.y,
			       best_l->end.x, best_l->end.y);
			/* Only increase speed on *first* bounce, to avoid
			 * mega speed increases. */
			bounce(&ball, best_l, &move, ignored ? 0.0 : 0.1);

			/* If we moved, stop ignoring lines. */
			if (ignored && (move.x > 0.001 || move.y > 0.001))
				clear_ignore(&lines);

			/* Don't hit the same line twice. */
			printf("Ignoring that line\n");
			best_l->ignore = ignored = true;
			goto again;
		}

		if (ignored && (move.x > 0.001 || move.y > 0.001)) {
			clear_ignore(&lines);
			printf("Moving by %f,%f to %f,%f\n",
			       move.x, move.y, new.x, new.y);
		}

		/* Restore what was there before ball. */
		if (savesur) {
			SDL_BlitSurface(savesur, NULL, screen, &rect);
			SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);
		} else {
			savesur = sub_surface(screen, ballsur->w, ballsur->h);
		}

		/* Save away image under new ball. */
		rect.w = ballsur->w;
		rect.h = ballsur->h;
		rect.x = new.x - ballsur->w/2;
		rect.y = new.y - ballsur->h/2;
		SDL_BlitSurface(screen, &rect, savesur, NULL);

		/* Draw new ball */
		SDL_BlitSurface(ballsur, NULL, screen, &rect);
		SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);
		/* That SDL_BlitSurface can crop rect. */
		rect.x = new.x - ballsur->w/2;
		rect.y = new.y - ballsur->h/2;
		ball.pos = new;
		SDL_Delay(25);

		if (SDL_PollEvent(&event)
		    && (event.type == SDL_QUIT
			|| (event.type == SDL_KEYDOWN
			    && event.key.keysym.sym == SDLK_ESCAPE))) {
			SDL_Quit();
			return 0;
		}

		if (cwiid_get_state(wiimote, &state))
			errx(1, "No wii state");

		/* Find biggest state. */
		ir = &state.ir_src[0];
		for (i = 0; i < CWIID_IR_SRC_COUNT; i++) {
			if (!state.ir_src[i].valid)
				continue;
			if (!ir->valid || state.ir_src[i].size > ir->size)
				ir = &state.ir_src[0];
		}

		if (ir->valid) {
			/* Give it some slack for missing one or two... */
			if (time_since_last_ir <= 5) {
				struct line_segment *n = malloc(sizeof(*n));
				/* Wiimote coordinates are backwards for us. */
				n->start.x = last_ir.pos[0];
				n->start.y = MAX_Y - last_ir.pos[1];
				n->end.x = ir->pos[0];
				n->end.y = MAX_Y - ir->pos[1];
				n->ignore = false;
				timeradd(&now, &line_life, &n->expires);
				list_add_tail(&lines, &n->list);
				thick_line(screen, n, 0);
			}
			time_since_last_ir = 0;
			last_ir = *ir;
		}
		time_since_last_ir++;
	}
}
